#!/usr/bin/env python
import re
import commands
import sys

def parseArgs(args):
  com=re.compile("^(.+)=(.+)$");
  opts={}
  for a in args:
    x=com.search(a);
    if x:
       opts[x.group(1)]=x.group(2);
  return opts;     


def dataSize(dat):
  data={}
  out=commands.getoutput("In3d "+dat);
  data["n1"]=re.compile("n1=(\S+)").search(out).group(1);
  data["n2"]=re.compile("n2=(\S+)").search(out).group(1);
  data["n3"]=re.compile("n3=(\S+)").search(out).group(1); 
  data["o1"]=re.compile("o1=(\S+)").search(out).group(1);
  data["o2"]=re.compile("o2=(\S+)").search(out).group(1);
  print "AA", re.compile("o2=(\S+)").search(out).group(1),"BB"
  data["o3"]=re.compile("o3=(\S+)").search(out).group(1); 
  data["d1"]=re.compile("d1=(\S+)").search(out).group(1);
  data["d2"]=re.compile("d2=(\S+)").search(out).group(1);
  data["d3"]=re.compile("d3=(\S+)").search(out).group(1); 
  return data

def docit():
  doc="Cubeplotsimp dat.H out.v \n";
  doc+="   point1,point2 = percentage of center frame 0-1\n";
  doc+="   frame1,frame2,frame3 - which frame to show\n";
  doc+="   center=1 - give center frame\n";
  doc+="   labelsz,labelfat\n"
  doc+="   geophysics=1\n"
  print doc
  sys.exit(-1);

def setDefaults(dat,args):
   if not args.has_key("point1"):   args["point1"]=0.5;
   if not args.has_key("point2"):   args["point2"]=0.5;
   if not args.has_key("frame1"):   args["frame1"]=float(dat["n1"])/2;
   if not args.has_key("frame2"):   args["frame2"]=float(dat["n2"])/2;
   if not args.has_key("frame3"):   args["frame3"]=float(dat["n3"])/2;
   if not args.has_key("crowd"):   args["crowd"]=.75;
   if not args.has_key("crowd1"):   args["crowd1"]=args["crowd"];
   if not args.has_key("crowd2"):   args["crowd2"]=args["crowd"];
   return args;

def calcDims(args):
   axis1Inch=13.65*args["crowd1"];
   axis2Inch=10.24*args["crowd2"];
   marg1=13.65-axis1Inch;
   marg2=10.24-axis2Inch;
   xll=marg1*2./3.
   yll=marg2*1./2.;
   xur=13.65-marg1/3.;
   yur=10.24-marg2/2.
   return xll,yll,xur,yur

def buildTop(dd,dat,args):
   out="tEmP_TOP.v";
   com="Window3d < %s f1=%d n1=1 |Grey out=ju_TOP.v geophysics=1 wantaxis2=n wanttitle=n >/dev/null"%(dat,int(args["frame1"]));
   commands.getoutput(com);
   xll,yll,xur,yur= calcDims(args);

   pct1=float(args["frame3"])/float(dd["n3"]);
   pct3=float(args["frame2"])/float(dd["n2"]);
   val1=float(dd["o2"])+float(dd["d2"])*int(args["frame2"])
   val2=float(dd["o3"])+float(dd["d3"])*int(args["frame3"])
   commands.getoutput("Overlay type=text color=6 x=%f y=%f  boxit=n out=l3.v label=%s"%((1.-pct1)*xll+pct3*xur,yur+.3,val2));
   commands.getoutput("Overlay type=text color=6 x=%f y=%f  boxit=n out=l4.v label=%s"%(xur+1.,(1.-pct1)*yll+pct3*yur,val1));
   
   commands.getoutput("Overlay type=polyline color=6 x=%f,%f y=%f,%f out=l1.v"%((1.-pct1)*xll+pct3*xur,
     (1.-pct1)*xll+pct3*xur,yll,yur))
   
   commands.getoutput("Overlay type=polyline color=6 x=%f,%f y=%f,%f out=l2.v"%(xll,xur,
     (1.-pct1)*yll+pct1*yur,
     (1.-pct1)*yll+pct1*yur));
   commands.getoutput("vppen  vpstyle=n erase=o ju_TOP.v l1.v l2.v l3.v l4.v >"+out);
   return out;

def buildSide(dd,dat,args):
   out="tEmP_SIDE.v";
   xll,yll,xur,yur= calcDims(args);
   com="Window3d < %s f2=%d n2=1 |Grey geophysics=1 out=ju_SIDE.v wanttitle=n wantaxis1=0>/dev/null"%(dat,int(args["frame2"]));
   commands.getoutput(com);
   val1=float(dd["o1"])+float(dd["d1"])*int(args["frame1"])
   val2=float(dd["o2"])+float(dd["d2"])*int(args["frame2"])
   pct1=float(args["frame1"])/float(dd["n1"]);
   pct3=float(args["frame3"])/float(dd["n3"]);
   commands.getoutput("Overlay type=text color=6 x=%f y=%f  boxit=n out=l3.v label=%s"%((1.-pct1)*xll+pct3*xur,yur+.3,val2));
   commands.getoutput("Overlay type=text color=6 x=%f y=%f  boxit=n out=l4.v label=%s"%(xur+1.,(1.-pct1)*yll+pct3*yur,val1));
   commands.getoutput("Overlay type=polyline color=6 x=%f,%f y=%f,%f out=l1.v"%((1.-pct1)*xll+pct3*xur,
     (1.-pct1)*xll+pct3*xur,yll,yur))
   
   commands.getoutput("Overlay type=polyline color=6 x=%f,%f y=%f,%f out=l2.v"%(xll,xur,
     (1.-pct1)*yll+pct1*yur,
     (1.-pct1)*yll+pct1*yur));
   commands.getoutput("vppen  vpstyle=n erase=o ju_SIDE.v l1.v l2.v l3.v l4.v >"+out);
   return out;


def buildFront(dd,dat,args):
   out="tEmP_FRONT.v";
   com="Window3d < %s f3=%d n3=1 |Grey out=ju_FRONT.v geophysics=1  wanttitle=n  >/dev/null"%(dat,int(args["frame3"]));
   commands.getoutput(com);
   xll,yll,xur,yur= calcDims(args);

   pct1=float(args["frame1"])/float(dd["n1"]);
   pct3=float(args["frame2"])/float(dd["n2"]);
   print commands.getoutput("Overlay type=polyline color=6 x=%f,%f y=%f,%f out=l1.v"%((1.-pct1)*xll+pct3*xur,
     (1.-pct1)*xll+pct3*xur,yll,yur))
   
   commands.getoutput("Overlay type=polyline color=6 x=%f,%f y=%f,%f out=l2.v"%(xll,xur,
     (1.-pct1)*yll+pct1*yur,
     (1.-pct1)*yll+pct1*yur));
   print commands.getoutput("vppen  vpstyle=n erase=o ju_FRONT.v l1.v l2.v >"+out);
   return out;

def buildPlot(args,out,top,side,front):
   xll,yll,xur,yur= calcDims(args);
   point2=0.5;point1=0.5;
   commands.getoutput("vppen %s vpstyle=n  xscale=%f yscale=%f >fR.v"%(front,point2,point1));
   commands.getoutput("vppen %s vpstyle=n xscale=%f yscale=%f xshift=%f >sI.v"%(side,1.-point2,point1,(xur-xll)*(point2)));
   commands.getoutput("vppen %s vpstyle=n xscale=%f yscale=%f yshift=%f>tO.v"%(top,point2,1.-point1,
    yur*point1-yll*(1.-point1)));
   print   yur*point1-yll*(1.-point1)
   commands.getoutput("vppen fR.v tO.v sI.v erase=o big=n vpstyle=n  >%s"%(out));


def buildOverlay():
   print "yeah"


def combineVplot():
 print "yeah"



if sys.argv < 4: 
  doc_it();
  sys.exit(-1)
data=dataSize(sys.argv[1]);
args=parseArgs(sys.argv[3:]);
args=setDefaults(data,args);
topFrame="tEmP_TOP.v"
sideFrame="tEmP_SIDE.v"
frontFrame="tEmP_FRONT.v"
topFrame=buildTop(data,sys.argv[1],args);
sideFrame=buildSide(data,sys.argv[1],args);
frontFrame=buildFront(data,sys.argv[1],args);
buildPlot(args,sys.argv[2],topFrame,sideFrame,frontFrame);





