#include "rtm_zero_op_3d.h"
#include "sinc_bob.h"
#include "gpu_funcs_3d.h"
#include <math.h>
#include "deriv_3d.h"
#include "laplac_3d.h"
#include "map_data_3d.h"
rtm_zero_op::rtm_zero_op(param_func *par,vel_fd_3d *vel_3d, source_func *source_func, data_rtm_3d *dat,data_rtm_3d *datin, image_rtm_3d *img,float ap,bool v,bool enc,int *r_vec, int n_zs, float p_zs,bool do_src,bool redo_src){
   std::vector<float> ds; 
   data=dat;
   bdata=datin;
   ds.push_back(dat->dt);
   ds.push_back(source_func->get_dt());
   set_vel(vel_3d);
   calc_stability(&ds,data->nt);
   set_fd_basics(par,source_func,ap,v);
   image=img;
   set_domain(image);
   set_range(data);
   dtd=data->dt;
   jtd=(int)(dtd/dt);
   nt=(data->nt-1)*jtd+1;
   base=0;
   redo=redo_src;
   create_transfer_sinc_data(8);

   encode=enc;
   rand_vec=r_vec;
   n_zeroes=n_zs;
   p_zeroes=p_zs;
   fprintf(stderr,"N_zs are %d, p_zs %f\n",n_zs,p_zs);

   if(do_src && !redo) create_source_fields();
}

void rtm_zero_op::create_transfer_sinc_data(int nsinc){
  jtd=(int)((data->dt)/dt);
  sinc_bob myr(jtd,nsinc);
  transfer_sinc_table_d(nsinc,jtd,myr.table);
}

bool rtm_zero_op::random_init(){
  int irand,count=0;
  int chk=1,nzs=0,irand2=0;
  bool fin=false;

  srand(time(NULL));
  int ns=data->nshots();
  while(fin==false){
    chk=0;
    nzs=0;
    for(int is=0; is<ns; is++){
      if(p_zeroes==0) {
        irand=rand()%2;
        if(irand==0) irand=-1;
      }
      else if(p_zeroes>66){
        irand2=rand()%6;
        if(irand2>4){ irand=1;}
        else if(irand2<1){  irand=-1;}
        else{  irand=0;}
      }
      else if(p_zeroes<34){
        irand2=rand()%6;
        if(irand2>3){  irand=1;}
        else if(irand2<2){  irand=-1;}
        else{  irand=0;}
      }
      else{
        irand2=rand()%4;
        if(irand2>2){  irand=1;}
        else if(irand2<1){  irand=-1;}
        else{  irand=0;}
      }

      rand_vec[is]=irand;
      chk=chk+irand;
      if(irand==0) nzs+=1;
    }
//fprintf(stderr,"reset %d ",nzs);
    if(abs(chk)<=1 && nzs==n_zeroes) fin=true;
  count+=1;
  }
  for(int is=0; is<ns; is++){
    fprintf(stderr,"%d ",rand_vec[is]);
  }
  fprintf(stderr," nzeros %d \n",nzs);

}

bool rtm_zero_op::data_init(my_vector *bdvec, int iter){
  data_rtm_3d *dn=(data_rtm_3d*) bdvec;
  bdata->set_source_file(dn);

  my_vector *dvec;
  //data_rtm_3d *d=(data_rtm_3d*) dvec;fprintf(stderr,"here2 \n");
  //data->set_source_file(d);fprintf(stderr,"here \n");
  bdvec->zero();

  int nt_big=8+data->nt;
  int npts=data->get_points();
  int npts1=data->get_points1();
  int npts2=data->get_points2();
  hypercube_float *src_p0=source->create_domain(1);
  float *s_z=new float[npts],*s_x=new float[npts],*s_y=new float[npts];
  axis ax_pt1(npts1);
  axis ax_pt2(npts2);

  time_t startdtimes, enddtimes;
  startdtimes = time(&startdtimes);

  float_3d *shot=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));

  data->get_source_func_encode(src_p0,0,rand_vec,s_x,s_y,s_z,8,nt_big,shot);
  enddtimes = time(&enddtimes);

  fprintf(stderr,"      Data init, in %f \n",difftime(enddtimes, startdtimes));

  bdata->add_data(0,shot);

  delete [] s_z;
  delete [] s_x;
  delete [] s_y;
  delete src_p0;
  delete shot;

  return true;
}

void rtm_zero_op::create_source_fields(){

  int nshots=data->nshots();
  auxinout("source_fields");

  int npts,nt_big;
  nt_big=9+data->nt; 
  npts=source->get_points(encode);
  int *locs=new int[npts];
  float *vals=new float[npts*(nt_big)];

  int seed=0;
  nshots=1;

  for(int ishot=0; ishot< nshots; ishot++){

     if(verb) fprintf(stderr,"Forward propagating shot %d of %d \n",ishot,nshots);
     hypercube_float *src_p0=source->create_domain(ishot);
     hypercube_float *src_p1=src_p0->clone();

     source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs,vals);

     transfer_source_func(npts,nt_big,locs,vals);
     seed=locs[3] - ((int) locs[3]/10000)*10000;

     hypercube_float *vloc=vel->rand_subspace(seed,nbound,nbound,nbound,dt,src_p0);

     transfer_vel_func1(nx,ny,nz,vloc->vals);

     time_t startwtimes, endwtimes;
     startwtimes = time(&startwtimes);

     source_prop(nx,ny,nz,false,true,src_p0->vals,src_p1->vals,jts,npts,nt);

     endwtimes = time(&endwtimes);

     fprintf(stderr,"Finished propagating wavefield, in %f \n",difftime(endwtimes, startwtimes));
     srite("source_fields",src_p0->vals,nz*nx*ny*4);
     srite("source_fields",src_p1->vals,nz*nx*ny*4);


     delete vloc;
     delete src_p0;
     delete src_p1;
   }
   delete [] locs;
   delete [] vals;
}

bool rtm_zero_op::forward(bool add, my_vector *mvec, my_vector *dvec, int iter){

  data_rtm_3d *bd=(data_rtm_3d*) dvec;
  image_rtm_3d *m=(image_rtm_3d*) mvec;

  if(!add) dvec->zero();

  int npts_s=source->get_points(encode);

  bdata->set_source_file(bd);
  image->set_source_file(m);

  int nt_big=8+data->nt;

  int npts=data->get_points();
  int rec_nx=data->get_points1();
  int rec_ny=data->get_points2();

  int *locs=new int[npts];
  int *locs_s=new int[npts_s];
  float *vals=new float[npts_s*(nt_big)];

  float *s_z=new float[npts],*s_x=new float[npts],*s_y=new float[npts];

  axis ax_pt1(rec_nx);
  axis ax_pt2(rec_ny);

  float_3d *shot=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));
  float_3d *shot_deriv=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));

  axis ax_t=axis(nt_big);
  float_3d *rec_func=new float_3d(ax_t,ax_pt1,ax_pt2);

  int nshots=data->nshots();

  nshots=1;
  for(int ishot=0; ishot<nshots; ishot++){

    if(verb) fprintf(stderr,"Forward shot %d of %d\n",ishot,data->nshots());
    hypercube_float *src_p0=source->create_domain(ishot);
    hypercube_float *src_p1=src_p0->clone();
    axis ad1=src_p0->get_axis(1);
    axis ad2=src_p0->get_axis(2);
    axis ad3=src_p0->get_axis(3);

    source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs_s,vals);

    transfer_source_func(npts_s,nt_big,locs_s,vals);
    int seed=locs_s[3] - ((int) locs_s[3]/10000)*10000;

    hypercube_float *vrand=vel->rand_subspace(seed,nbound,nbound,nbound,dt,src_p0);
    hypercube_float *vnone=vel->rand_subspace(seed,0,0,0,dt,src_p0);
    hypercube_float *img=image->extract_sub(ad1,ad2,ad3);
    transfer_vel_func1(nx,ny,nz,vrand->vals);
    transfer_vel_func2(nx,ny,nz,vnone->vals);

    bdata->get_source_func(src_p0,ishot,s_x,s_y,s_z,8,nt_big,shot);

    map_data mapit(npts,s_x,s_y,s_z,locs,src_p0,rec_func,shot_deriv,nt_big);

    //rec_func->zero();

    transfer_receiver_func(rec_nx,rec_ny,nt_big,locs,rec_func->vals/*,s_z,s_x,s_y*/);

    //hypercube_float *img_lap=img->clone();
    //laplac lap=laplac(img_lap,img);
    //lap.forward(false,img_lap,img);

    //lap.dot_test(true);

    //fprintf(stderr,"Data dims %d %d %d \n",ax_t.n,ax_pt1.n,ax_pt2.n);
    //fprintf(stderr,"Data dims %d %d %d \n",data->get_axis(1).n,data->get_axis(2).n,data->get_axis(3).n);

    rtm_forward(ad1.n,ad2.n,ad3.n,jtd,img->vals,rec_func->vals,npts_s,nt,nt_big,rec_nx,rec_ny);
    //rtm_forward(ad1.n,ad2.n,ad3.n,jtd,img->vals,rec_func->vals,npts_s,nt,nt_big,rec_nx,rec_ny);

    deriv der=deriv(shot_deriv,shot);

    mapit.forward(false,rec_func,shot_deriv);

    der.forward(false,shot_deriv,shot);

    bdata->add_data(0,shot);


    delete src_p0; delete src_p1; delete img;// delete img_lap;
    delete vrand; delete vnone;
  }
  delete [] s_z;
  delete [] s_x;
  delete [] s_y;
  delete [] locs_s;
  delete [] vals;
  delete [] locs;
  delete rec_func;
  delete shot;
  delete shot_deriv;
  return true;

}

bool rtm_zero_op::adjoint(bool add, my_vector *mvec, my_vector *dvec, int iter){

  if(redo){
     create_source_fields();
  }
  if(!add) {
    mvec->zero();
  }

  data_rtm_3d *bd=(data_rtm_3d*) dvec;
  image_rtm_3d *m=(image_rtm_3d*) mvec;
  int npts_s=source->get_points(encode);
  bdata->set_source_file(bd);

  image->set_source_file(m);
  int nt_big=8+data->nt;
  int npts=data->get_points();
  int rec_nx=data->get_points1();
  int rec_ny=data->get_points2();

  int *locs=new int[npts];
  int *locs_s=new int[npts_s];
  float *vals=new float[npts_s*(nt_big)];

  float *s_z=new float[npts],*s_x=new float[npts],*s_y=new float[npts];
  axis ax_pt1(rec_nx);
  axis ax_pt2(rec_ny);

  float_3d *shot=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));
  float_3d *shot_deriv=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));
  axis ax_t=axis(nt_big);

  float_3d *rec_func=new float_3d(ax_t,ax_pt1,ax_pt2);
  int seed=0;

  sseek("source_fields",0,0);
  int nts=(int)(nt/jts)+1;
  int nshots=data->nshots();

  nshots=1;

  for(int ishot=0; ishot<nshots; ishot++){
    if(verb) fprintf(stderr,"Adjoint shot %d of %d \n",ishot,data->nshots());
    hypercube_float *src_p0=source->create_domain(ishot);
    hypercube_float *src_p1=src_p0->clone();
    axis ad1=src_p0->get_axis(1);
    axis ad2=src_p0->get_axis(2);
    axis ad3=src_p0->get_axis(3);

    source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs_s,vals);
    bdata->get_source_func(src_p0,ishot,s_x,s_y,s_z,8,nt_big,shot);

    transfer_source_func(npts_s,nt_big,locs_s,vals);

    seed=locs_s[3] - ((int) locs_s[3]/10000)*10000;

    hypercube_float *vrand=vel->rand_subspace(seed,nbound,nbound,nbound,dt,src_p0);
    //hypercube_float *vnone=vel->rand_subspace(seed,0,0,0,dt,src_p0);
    hypercube_float *img=src_p0->clone();

    transfer_vel_func1(nx,ny,nz,vrand->vals);
    //transfer_vel_func2(nx,ny,nz,vnone->vals);

    deriv der=deriv(shot_deriv,shot);
    der.adjoint(false,shot_deriv,shot);

    map_data mapit(npts,s_x,s_y,s_z,locs,src_p0,rec_func,shot_deriv,nt_big);
    mapit.adjoint(false,rec_func,shot_deriv);

    sreed("source_fields",src_p1->vals,4*ad1.n*ad2.n*ad3.n);
    sreed("source_fields",src_p0->vals,4*ad1.n*ad2.n*ad3.n);

    transfer_receiver_func(rec_nx,rec_ny,nt_big,locs,rec_func->vals/*,s_z,s_x,s_y*/);

    time_t startwtimem, endwtimem;
    startwtimem = time(&startwtimem);

    rtm_adjoint(ad1.n,ad2.n,ad3.n,jtd,src_p0->vals,src_p1->vals,img->vals,npts_s,nt/*,src,recx*/);

    endwtimem = time(&endwtimem);
    //fprintf(stderr,"Data dims %d %d %d \n",ad1.n,ad2.n,ad3.n);

    fprintf(stderr,"Adjoint, wall clock time = %f \n", difftime(endwtimem, startwtimem));

    //hypercube_float *img_lap=img->clone();
    //laplac lap=laplac(img_lap,img);
    //lap.adjoint(false,img_lap,img);
    //lap.dot_test(true);

  /*hypercube_float *mute=(hypercube_float*)image->clone_zero();
  sreed("mute",mute->vals,4*a1o.n*a2o.n*a3o.n);
  sseek("mute",0,0);
  for(int h=0; h<image->get_n123(); h++){
    image->vals[h]=image->vals[h]*mute->vals[h];
  }*/

    image->add_image(img);
    //image->add_image(img_lap);

    delete src_p0; delete src_p1; delete img; //delete img_lap;
    delete vrand;
  }
  sseek("source_fields",0,0);

  image->write_volume();

  delete [] s_z;
  delete [] s_x;
  delete [] s_y;
  delete [] locs_s;
  delete [] vals;
  delete [] locs;
  delete rec_func;
  delete shot;
  delete shot_deriv;

  return true;
}
