module ADCIG_mod

  use sep
  use fourier_mod
  use smooth3d_mod

  implicit none
  integer         , private         :: nx, np, nt
  real            , private         :: ox, dx, op, dp, ot, dt,of,df
  real            , private         :: knyq, fnyq
  real, allocatable, dimension(:,:)              :: filterArr

  integer         , private         :: nnx, nnp, nnt
  real            , private         :: oox,oop,ddx,ddp,oot,ddt


contains

  subroutine adcig_init(nx_in,ox_in,dx_in,np_in,op_in,dp_in, nt_in, ot_in, dt_in)
    integer              :: nx_in, np_in, nt_in
    real                 :: ox_in, dx_in, op_in, dp_in, ot_in, dt_in
    nx = nx_in
    np = np_in
    ox = ox_in
    op = op_in
    dx = dx_in
    dp = dp_in
    nt = nt_in
    ot = ot_in
    dt = dt_in

    write(0,*) 'this is the end of adcig_init'

    end subroutine



!!--------------------------------------------------------------------------
!! forward: (t,x) to (tau,p) and backward (tau,p) to (t,x)
!! Thanks Gboyega for the code
!! data: 2D array in tau-p domain
!! modl: 2D array in t-x domain;
!! adj: (0) from modl to data   (1) from data to modl 

  subroutine adcig_oper(adj, add, modl, data)

   logical                  :: adj, add
   real, dimension(:,:)     :: modl, data   
   integer                 :: counter, indm, indd
   integer                 :: it, ix, ip, itau
   integer                 :: imin, imax, ishift
   real                    :: p, tau, x, oox, xmid, xrange
   real                    :: shift, samp_shift, del,signs


!   write(0,*) '**************************************'
!   write(0,*) 'in adcig_oper'
    

    if (.not. add) then
       if (adj) then
          modl = 0.
       else
          data = 0.
       end if
    end if


  if (adj) then
     modl=0.
  else
     data=0.
  endif

  counter=0
  xmid = (nx-1)/2+1

!   write(0,*) 'np,op,dp',np,op,dp
!   write(0,*) 'nx,ox,dx',nx,ox,dx
!   write(0,*) 'nt,ot,dt',nt,ot,dt
!write(0,*) 'size data', size(data,1), size(data,2)
!write(0,*) 'size model', size(modl,1), size(modl,2)

   do itau=1,nt
      tau = ot+(itau-1)*dt

      ! xrange increases with depth (10 to 25 gridcell)
      xrange =  10 + int((itau/nt)*15)
 
 !      do ip=31,31     
      do ip=1,np
          p = op+(ip-1)*dp
!!          write(0,*) 'ip, p,itau', ip, p, itau

!        do ix=xmid,xmid
!        do ix=xmid-xrange,xmid+xrange
         do ix=1,nx
           x = ox+(ix-1)*dx
!!           write(0,*) 'xmid,ix,x',xmid,ix,x 
         
           shift= p*x
           samp_shift= shift/dt
           if (samp_shift .ne. 0.) then
              signs = samp_shift/abs(samp_shift)
           else 
              signs=0
           end if
           ishift=signs*floor(abs(samp_shift))
           del = samp_shift-ishift
          
!!           write(0,*) '!!', signs, floor(abs(samp_shift))
!!           write(0,*) '@@',x,p, shift,samp_shift,ishift

           it = itau+ishift 
          
           if ((it .lt. 1) .or. (it .gt. nt)) then
              counter = counter+1
              cycle
           end if

            if (adj) then     ! adjoint --> with linear interp
               modl(it,ix) =  modl(it,ix) + data(itau,ip)*(1.-del) &
                      + data(itau+1,ip)*del     
            else               ! Forward -> linear interpT
!               data(itau,ip) =  data(itau,ip) + modl(it,ix);
  !!             write(0,*) 'itau,ip',itau,ip,it,ix
  !!             call erexit('exit now')
               data(itau,ip) =  data(itau,ip) + modl(it,ix)*(1.-del);
               data(itau+1,ip) =  data(itau+1,ip) + modl(it,ix)*del; 
            endif

        enddo
      enddo
    enddo

!    write(0,*) counter, 'out of', np*nx, 'excluded', xmid

 !   call erexit('exit now')

  end subroutine

!------------------------------------------------------------
! Mandy's version of tau-p transform


  subroutine adcig2_oper(adj, add, modl, data) 

   logical                  :: adj, add
   real, dimension(:,:)     :: modl, data
   integer                 :: counter
   integer                 :: it, ix, ip, itau
   integer                 :: imin, imax, ishift
   real                    :: p, tau, x, oox
   real                    :: tloc, del, t, shift, samp_shift

    


  write(0,*) '**************************************'
  write(0,*) 'in Mandy adcig_oper'
  write(0,*) 'adj, add', adj, add
  write(0,*) 'nx, np, nt', nx, np ,nt
  write(0,*) '**************************************'

  if (adj) then
     modl=0.
  else
     data=0.
  endif

  counter=0
  ! loop over offsets
    do ix=1,nx 
       x= (ox + (ix-1)*dx)
       do ip=1,np  ! loop over p 
          p = op+(ip-1)*dp  

          shift= -p*x
          samp_shift= shift/dt

          ishift=floor(samp_shift)
          imin=max( 1, -ishift+1)
          imax=min(nt-1, nt-ishift)
          del=samp_shift-ishift
          itau=imin+ishift
             
          if(imin>nt) then 
!              write(0,*) 'case when imin > nt'
	      counter=counter+1
              cycle
          end if

          do it=imin,imax;  ! loop over time

             if (adj) then     ! adjoint --> with linear interp
                modl(it,ix) =  modl(it,ix) + data(itau,ip)*(1.-del) &
                     + data(itau+1,ip)*del     
             else               ! Forward -> linear interpT
                data(itau,ip) =  data(itau,ip) + modl(it,ix)*(1.-del);
                data(itau+1,ip) =  data(itau+1,ip) + modl(it,ix)*del; 
             endif

             itau=itau+1
          
        enddo
	
	
       enddo
    enddo

	write(0,*) counter, 'out of', np*nx*nt, 'excluded'

  end subroutine

!--------------------------------------------------------------------------------
! Anti-aliasing filter for p-tau trasform (Dev and McMechan 2009)
! Filter is appied in p-f domain with the shape  k_nyquist = |px-pi|f
! step 1: Apply 1D Fourier transform from tau to f (p-tau) to (p-f)
! step 2: apply filter (note that filter is symmetric about pi and about 
!         half of the time-domain nyquist frequency. 
!        For this study, will choose pi to be zero
!        this assumes that most of the dip is near horizontal (which is kind of true) 
! step 3: Inverse 1D Fourier transform back to (p-tau)
!--------------------------------------------------------------------------------

  subroutine  DealiasPTAU_init( nx, dx, nt,dt)

      integer   :: iif,nf,ip,nx,nt
      real :: dx, dt
      real :: p, f
      real :: pi=4.*atan(1.)


   ! calculate wavenumber nyquist and time-domain nyquist
     knyq = 1/(2.*dx)
     fnyq = 1/(2.*dt)
     nf = nt
     of = -1./(2*dt)
     df = 1./(nt*dt)
     

     write(0,*) 'nx,dx',nx,dx
     write(0,*) 'nt,dt',nt,dt
     write(0,*) 'nf,of,df',nf,of,df
     write(0,*) 'np,op,dp',np,op,dp
     write(0,*) 'knyq', knyq
     
    ! Define the filter here
    allocate(filterArr(nt,np))

    ! initialize to zero first 
    filterArr = 0.
    call ft_init(nt, np)

    do ip=1,np
        p = op+(ip-1)*dp  
       do iif=1,nf
            f = of + (iif-1)*df
            if (abs(p*f) .lt. knyq/3.) then !set to 1 
                filterArr(iif,ip) = 1.          
            end if
       end do
    end do



    ! smooth it a little bit to create tapering affects
    call triangle2( 5, 5, nt, np, filterArr, filterArr)
         
     call sep_write(filterArr,"filterArr.H")
   call auxputch("n1",'i', nf, "filterArr.H");      call auxputch("label1",'s', 'f', "filterArr.H"); 
   call auxputch("d1",'f', df, "filterArr.H");      call auxputch("o1",'f', of, "filterArr.H");
   call auxputch("n2",'i', np, "filterArr.H");      call auxputch("label2",'s', 'x', "filterArr.H"); 
   call auxputch("d2",'f', dp, "filterArr.H");      call auxputch("o2",'f', op, "filterArr.H");
   
 !  call erexit('exit now')

  end subroutine 

!--------------------------------------------------------------------------------


  subroutine DealiasPTAU_oper(input,output)

     integer                  :: ip, it
     real                     :: sign1
     real, dimension(:,:)     :: input, output
     complex, allocatable, dimension(:,:)   :: cdata,pfArr

     allocate(pfArr(nt,np), cdata(nt,np))

     do ip=1,np
        do it=1,nt
           cdata(it,ip) = cmplx(input(it,ip))
        end do
     end do

     ! Forward 1D Fourier transform
       sign1=1.
       call ft1axis(.false., sign1, cdata)

   if (3 > 5) then
   call sep_write(cdata,"cdata.H")
   call auxputch("n1",'i', nt, "cdata.H");      call auxputch("label1",'s', 'f', "cdata.H"); 
   call auxputch("d1",'f', df, "cdata.H");      call auxputch("o1",'f', of, "cdata.H");
   call auxputch("n2",'i', np, "cdata.H");      call auxputch("label2",'s', 'x', "cdata.H"); 
   call auxputch("d2",'f', dp, "cdata.H");      call auxputch("o2",'f', op, "cdata.H");
!   call erexit('exit now')
   end if

     ! Apply filter
     do ip=1,np
        do it=1,nt
!           pfArr(it,ip) = cdata(it,ip)*filterArr(it,ip)
	    pfArr(it,ip) = cdata(it,ip)
        end do
     end do


     ! Inverse 1D Fourier transform
       call ft1axis(.true., sign1, pfArr)       
       

     do ip=1,np
        do it=1,nt
           output(it,ip) = real(pfArr(it,ip))
        end do
     end do


  end subroutine

!--------------------------------------------------------------------------------

  subroutine Anglefilter_init(nnx_in,oox_in,ddx_in,nnp_in,oop_in,ddp_in, nnt_in, oot_in, ddt_in)
    integer              :: nnx_in, nnp_in, nnt_in
    real                 :: oox_in, ddx_in, oop_in, ddp_in, oot_in, ddt_in
    nnx = nnx_in
    nnp = nnp_in
    oox = oox_in
    oop = oop_in
    ddx = ddx_in
    ddp = ddp_in
    nnt = nnt_in
    oot = oot_in
    ddt = ddt_in

    write(0,*) 'this is the end of Anglefilter_init'

    end subroutine
!--------------------------------------------------------------------------------
subroutine angle_taup(adj, add, modl, data) 

   logical                  :: adj, add, verb
   real, dimension(:,:)     :: modl, data
   integer                 :: counter
   integer                 :: it, ix, ip, itau
   integer                 :: imin, imax, ishift
   real                    :: p, tau, x
   real                    :: tloc, del, t, shift, samp_shift

  verb=.true.

  if (verb) then
  write(0,*) '**************************************'
  write(0,*) 'in Mandy angle_taup'
  write(0,*) 'adj, add', adj, add
  write(0,*) 'nnx, oox, ddx', nnx, oox, ddx
  write(0,*) 'nnp, oop, ddp', nnp, oop, ddp
  write(0,*) 'nnt, oot, ddt', nnt, oot, ddt
  write(0,*) 'size model', size(modl,1), size(modl,2)
  write(0,*) 'size data', size(data,1), size(data,2)
  write(0,*) '**************************************'
  end if


  if (adj) then
     modl=0.
  else
     data=0.
  endif

  counter=0
  ! loop over offsets
    do ix=1,nnx 
       x= (oox + (ix-1)*ddx)
       do ip=1,nnp  ! loop over p 
          p = oop+(ip-1)*ddp  

          shift= -p*x
          samp_shift= shift/ddt

          ishift=floor(samp_shift)
          imin=max( 1, -ishift+1)
          imax=min(nnt-1, nnt-ishift)
          del=samp_shift-ishift
          itau=imin+ishift

          write(0,*) 'ix,ip',ix,ip,imin,imax
             
          if ((imin>nnt) .or. (imin > imax)) then 
!              write(0,*) 'case when imin > nt'
	      counter=counter+1
              cycle
          end if

          do it=imin,imax;  ! loop over time

             if (adj) then     ! adjoint --> with linear interp
                modl(it,ix) =  modl(it,ix) + data(itau,ip)*(1.-del) &
                     + data(itau+1,ip)*del     
             else               ! Forward -> linear interpT
                data(itau,ip) =  data(itau,ip) + modl(it,ix)*(1.-del);
                data(itau+1,ip) =  data(itau+1,ip) + modl(it,ix)*del; 
             endif

             itau=itau+1
          
        enddo
	
	
       enddo
    enddo

    if (verb) write(0,*) counter, 'out of', nnp*nnx, 'excluded'
   call erexit('exit now')

  end subroutine

!--------------------------------------------------------------------------------
! filters out all the non-flat component in ADCIG
! (1): linear tau-p transformon the angle domain
! (2): filter out all the p value greater than certain limit

  subroutine AngleDomain_filter(n2p,npmid,n1tau,input)

    real, dimension(:,:) :: input
    integer              :: ip, ipb,itau, n2p, n1tau,npmid
    real, allocatable, dimension(:,:)  :: weight

    allocate(weight(n1tau,n2p))
    
    weight = 0.

        do itau=1,n1tau
              if (itau .gt. 0.25*n1tau) then
                   ipb = int(((itau-0.25*n1tau)*10)/(0.75*n1tau))+3
              else
                   ipb = 3
              end if
              weight(itau,npmid+1-ipb:npmid+1+ipb) = 1.
        end do
                  
    ! smooth it a little bit to create tapering affects
    call triangle2( 3, 3, nt, np, weight, weight)
       
    if (1>2) then
        call sep_write(weight,"wgtfile.H")
        call auxputch("n1",'i', n1tau, "wgtfile.H");      call auxputch("label1",'s', 'z', "wgtfile.H"); 
        call auxputch("d1",'f', dt, "wgtfile.H");      call auxputch("o1",'f', ot, "wgtfile.H");
        call auxputch("n2",'i', n2p, "wgtfile.H");      call auxputch("label1",'s', 'p', "wgtfile.H"); 
        call auxputch("d2",'f', dp, "wgtfile.H");      call auxputch("o1",'f', op, "wgtfile.H");
        call erexit('exit now to view wgitfile')
    end if

    do ip=1,n2p
       do itau=1,n1tau
          input(itau,ip) = input(itau,ip)*weight(itau,ip)
       end do
    end do  


  end subroutine
!--------------------------------------------------------------------------------


end module 
