! Apply NMO operator
program NMO_operator3d

!Transpose image first!!
  use sep
  use Hradon_mod
  use omp_lib
  use mpi

  implicit none

  integer                           :: nz, nhx, nzeta, np, n3, i3
  real                              :: oz,dz,ohx,dhx,ozeta,dzeta,op,dp,o3,d3
  logical                           :: adj
  real, allocatable, dimension(:,:,:) :: model, data

  integer :: inode, node, nodes



  call sep_init()

  call from_param("nodes" ,nodes, 1)
  node =omp_get_num_procs()
  nodes=min(node-1,nodes)
  write(0,*) "Number of nodes available",node,"using",nodes

  call from_param("adj", adj)

  if(adj) then
    call from_history("n1", nz)
    call from_history("o1", oz)
    call from_history("d1", dz)
    call from_history("n2", nhx)
    call from_history("o2", ohx)
    call from_history("d2", dhx)
    call from_history("n3", n3)
    call from_history("o3", o3)
    call from_history("d3", d3)
    call from_param("nzeta", nzeta)
    call from_param("ozeta", ozeta)
    call from_param("dzeta", dzeta)
    call from_param("np", np)
    call from_param("op", op)
    call from_param("dp", dp)
    call to_history("n1", nzeta)
    call to_history("o1", ozeta)
    call to_history("d1", dzeta)
    call to_history("n2", np)
    call to_history("o2", op)
    call to_history("d2", dp)
    call to_history("n3", n3)
    call to_history("o3", o3)
    call to_history("d3", d3)
    call to_history("label1", "zeta")
    call to_history("label2", "p")
    call to_history("label3", "x")
  else
    call from_history("n1", nzeta)
    call from_history("o1", ozeta)
    call from_history("d1", dzeta)
    call from_history("n2", np)
    call from_history("o2", op)
    call from_history("d2", dp)
    call from_history("n3", n3)
    call from_history("o3", o3)
    call from_history("d3", d3)
    call from_param("nz", nz)
    call from_param("oz", oz)
    call from_param("dz", dz)
    call from_param("nhx", nhx)
    call from_param("ohx", ohx)
    call from_param("dhx", dhx)
    call to_history("n1", nz)
    call to_history("o1", oz)
    call to_history("d1", dz)
    call to_history("n2", nhx)
    call to_history("o2", ohx)
    call to_history("d2", dhx)
    call to_history("n3", n3)
    call to_history("o3", o3)
    call to_history("d3", d3)
    call to_history("label1", "z")
    call to_history("label2", "hx")
    call to_history("label3", "x")
  end if

  allocate(model(nzeta, np,n3))
  allocate(data(nz, nhx,n3))
  model = 0.
  data = 0.

  call omp_set_num_threads(nodes); 

  write(0,*) "Image should be transposed to z,hx,x..."

  call nmo_init(nzeta, ozeta, dzeta, np, op, dp, nz, oz, dz, nhx, ohx, dhx)

!$OMP PARALLEL 
inode=omp_get_thread_num()
write(0,*) "Hi from node",inode
!$OMP END PARALLEL


  if(adj) then
    call sep_read(data)
!$OMP PARALLEL DO private(inode,i3) SHARED(data,model) num_threads(nodes) 
    do i3=1,n3
      call nmo_op2(adj,.false.,model(:,:,i3),data(:,:,i3))
      !write(0,*) "Finished ",i3,"/",n3
    end do
!$OMP END PARALLEL DO
    call sep_write(model)
  else
    call sep_read(model)

!$OMP PARALLEL DO private(inode,i3) SHARED(data,model) num_threads(nodes) 
    do i3=1,n3
      call nmo_op2(adj,.false.,model(:,:,i3),data(:,:,i3))
      inode=omp_get_thread_num()
      !write(0,*) "Finished ",i3,"/",n3,"on node",inode
    end do
!$OMP END PARALLEL DO
    call sep_write(data)
  end if


  call sep_close()

end program
