program Pradon_3d_inv2

!Transpose image first!!
  use sep
  use Pradon_mod
  use omp_lib

  implicit none

  integer                           :: nz, nhx, nzeta, np, n3, i3
  real                              :: oz,dz,ohx,dhx,ozeta,dzeta,op,dp,o3,d3
  real, allocatable, dimension(:,:,:) :: model, data

  real, allocatable, dimension(:,:) :: rr,gg,m,g,s,ss
  double precision :: alfa,beta,gdg,sds,gds,determ,gdr,sdr
  integer :: iter,niter
  logical :: forget

  integer :: inode, node, nodes


  call sep_init()

  call from_param("nodes" ,nodes, 1)
  node =omp_get_num_procs()
  nodes=min(node-1,nodes)
  write(0,*) "Number of nodes available",node,"using",nodes

  call from_param("niter", niter,10)

  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nhx)
  call from_history("o2", ohx)
  call from_history("d2", dhx)
  call from_history("n3", n3)
  call from_history("o3", o3)
  call from_history("d3", d3)
  call from_param("nzeta", nzeta)
  call from_param("ozeta", ozeta)
  call from_param("dzeta", dzeta)
  call from_param("np", np)
  call from_param("op", op)
  call from_param("dp", dp)
  call to_history("n1", nzeta)
  call to_history("o1", ozeta)
  call to_history("d1", dzeta)
  call to_history("n2", np)
  call to_history("o2", op)
  call to_history("d2", dp)
  call to_history("n3", n3)
  call to_history("o3", o3)
  call to_history("d3", d3)
  call to_history("label1", "zeta")
  call to_history("label2", "p")
  call to_history("label3", "x")

  allocate(model(nzeta, np,n3))
  allocate(data(nz, nhx,n3))
  model = 0.
  data = 0.

  call omp_set_num_threads(nodes); 

  write(0,*) "Image should be transposed to z,hx,x..."

  call pradon_init(nzeta, ozeta, dzeta, np, op, dp, nz, oz, dz, nhx, ohx, dhx)

!$OMP PARALLEL 
inode=omp_get_thread_num()
write(0,*) "Hi from core",inode
!$OMP END PARALLEL


  call sep_read(data)

  data=data/maxval(data)

!$OMP PARALLEL DO ordered private(inode,i3,rr,gg,ss,m,g,s,alfa,beta,gdg,sds,gds,gdr,sdr,forget) SHARED(data,model) num_threads(nodes) schedule(dynamic,1)
  do i3=1,n3
    if(allocated(rr)) deallocate(rr)
    if(allocated(gg)) deallocate(gg)
    if(allocated(ss)) deallocate(ss)
    if(allocated(m)) deallocate(m)
    if(allocated(g)) deallocate(g)
    if(allocated(s)) deallocate(s)

    allocate(rr(nz, nhx))
    allocate(gg(nz, nhx))
    allocate(ss(nz, nhx))

    allocate(m(nzeta, np))
    allocate(g(nzeta, np))
    allocate(s(nzeta, np))

    rr=0;m=0;g=0;gg=0
    rr(:,:)=data(:,:,i3)
    forget=.true.
    do iter=1,niter
      call pradon_op2(.true.,.false.,g,rr)   !g=F'Rd
      call pradon_op2(.false.,.false.,g,gg)  !Gd=Fg

      if ( forget) then  !x, g, rr, gg = m,g,rr,gg
        s = 0.
        ss = 0.
        beta = 0.d0    ! steepest descent
        if ( sum(dprod(gg, gg)) .eq. 0 ) then
          call erexit('cgstep: grad vanishes identically')
        end if
        alfa = - sum( dprod( gg, rr)) / sum( dprod( gg, gg))
      else
        gdg = sum( dprod( gg, gg))   ! search plane by solving 2-by-2
        sds = sum( dprod( ss, ss))   !  G . (R - G*alfa - S*beta) = 0
        gds = sum( dprod( gg, ss))   !  S . (R - G*alfa - S*beta) = 0
        if ( gdg.eq.0. .or. sds.eq.0.) then
          write(0,*) "We gots a problem"
        end if
        determ = gdg * sds * max (1.d0 - (gds/gdg)*(gds/sds), 1.d-12)
        gdr = - sum( dprod( gg, rr))
        sdr = - sum( dprod( ss, rr))
        alfa = ( sds * gdr - gds * sdr ) / determ
        beta = (-gds * gdr + gdg * sdr ) / determ
	!write(0,*) "Iter",iter,gdg,sds,gds,gdr,sdr,alfa,beta
      end if 
      s  = alfa *  g + beta *  s            ! update solution step
      ss = alfa * gg + beta * ss            ! update residual step
      m  =  m +  s                          ! update solution
      rr = rr + ss                          ! update residual
      forget = .false.
      !write(0,*) "Finished ",i3,"/",n3,maxval(rr),maxval(g),maxval(gg),maxval(s)

    end do

    if (modulo(i3,n3/10) .eq. 0) then
       write(0,*) i3/n3,"% completed"
    end if

    model(:,:,i3)=m
  end do
!$OMP END PARALLEL DO

  call sep_write(model)
  call sep_close()

end program
