module fourier_mod

  implicit none
  include "/opt/FFTW/include/fftw3.f"
!./intel2/mkl/10.2.5.035/include/fftw/fftw3.f
!./FFTW/fftw-3.3/api/fftw3.f
!./FFTW/3.3_single/include/fftw3.f
!./FFTW/fftw-3.1.2/api/fftw3.f
!./FFTW/include/fftw3.f
!./Intel/Compiler/11.1/059/mkl/include/fftw/fftw3.f
!./Intel/mkl/10.2.5.035/include/fftw/fftw3.f
!./Intel/composerxe-2011.5.220/mkl/include/fftw/fftw3.f
!./Intel/composerxe-2011.0.084/mkl/include/fftw/fftw3.f
!./Intel/composerxe-2011.1.107/mkl/include/fftw/fftw3.f
!./FFTW_old/fftw-3.1.2/api/fftw3.f
!./FFTW_old/include/fftw3.f


  complex, dimension(:), allocatable, private  :: in1, out1, in2, out2
  integer, private                             :: n1, n2
  integer*8, private                           :: plan1f, plan1b, plan2f, plan2b
  contains

  subroutine ft_init(n1_in, n2_in)
    integer               :: n1_in, n2_in
    n1 = n1_in
    n2 = n2_in
    allocate(in1(n1), out1(n1))
    allocate(in2(n2), out2(n2))
    in1 = 0.
    out1 = 0.
    in2 = 0.
    out2 = 0.
    call sfftw_plan_dft_1d(plan1f, n1, in1, out1, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan1b, n1, out1, in1, FFTW_BACKWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan2f, n2, in2, out2, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan2b, n2, out2, in2, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine

  subroutine ft1axis(adj, sign1, cdata)
    logical, intent(in)        :: adj
    integer                    :: i2
    complex, dimension(:,:) :: cdata
    real                       :: sign1
    if(adj) then
      do i2=1,n2
        out1(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1,i2)
        out1((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2,i2)
        call sfftw_execute_dft(plan1b, out1, in1)
        cdata(:,i2) = in1
      enddo
    else
      do i2=1,n2
        in1 = cdata(:,i2)
        call sfftw_execute_dft(plan1f, in1, out1)
        cdata(:,i2) = out1
        cdata(1:n1-(n1+1)/2,i2) = out1((n1+1)/2+1:n1)
        cdata(n1-(n1+1)/2+1:n1,i2) = out1(1:(n1+1)/2)
      enddo
    endif
    cdata = sign1*cdata/sqrt(1.*n1)
  end subroutine

  subroutine ft2axis(adj, sign2, cdata)
    logical, intent(in)        :: adj
    integer                    :: i1
    complex, dimension(:,:) :: cdata
    real                       :: sign2
    if(adj) then
      do i1=1,n1
        out2(1:(n2+1)/2) = cdata(i1,n2-(n2+1)/2+1:n2)
        out2((n2+1)/2+1:n2) = cdata(i1,1:n2-(n2+1)/2)
        call sfftw_execute_dft(plan2b, out2, in2)
        cdata(i1,:) = in2
      enddo
    else
      do i1=1,n1
        in2 = cdata(i1,:)
        call sfftw_execute_dft(plan2f, in2, out2)
        cdata(i1,:) = sign2*out2/sqrt(1.*n2)
        cdata(i1,1:n2-(n2+1)/2) = out2((n2+1)/2+1:n2)
        cdata(i1,n2-(n2+1)/2+1:n2) = out2(1:(n2+1)/2)
      enddo
    endif
    cdata = sign2*cdata/sqrt(1.*n2)
  end subroutine

end module
