program myOff2Ang2D_back
  ! Convert ODCIT to ADCIG
  ! Input: Offset-domain CIG 
  !      : axis 1=z, 2=x 3=y, 4=hx
  ! Output: Angle-domain CIG
  !      : axis 1=z, 2=x 3=y, 4=ax 

  use sep
  use omp_lib
 ! use rtm3d_operAngle
  use ADCIG_mod
 ! use waveprop3d_mod
 ! use type3d_mod
  use ifport

  use solver_mod
  use cgstep_mod

  implicit none

  logical(4) :: success      
  logical                   :: verb
  integer                   :: it, nt, n4, niter
  integer                   :: iz, nz 
  integer                   :: ix, nx
  integer                   :: iy, ny
  integer                   :: ihx, nhx, iax, nax
  real                      :: dax, oax, dhx, ohx
  real                      :: dx, dz, dy, ox, oz, oy
  integer                   :: jsnap
  real, dimension(:,:), allocatable  :: tmpZX, tmpZA, tmpZA2, tmpZA3, dealiasZA
  real, dimension(:,:,:,:), allocatable  :: model, model2
  character(len=128)        :: out_

  ! OMP stuff 
  integer :: inode,node,nodes,Totalnodes


  call sep_init()


  !!======================================================
  !! Parameter reading
  
  write(0,*) ' This is the beginning of myOff2Ang'

   call from_param("verb",verb,.false.);      


  !! input file (offset domain CIG)
   call from_history("n1", nz);
   call from_history("d1", dz);
   call from_history("o1", oz);
   call from_history("n2", nx);
   call from_history("d2", dx);
   call from_history("o2", ox);
   call from_history("n3", n4);
   nax = (n4-1)/2 
   call from_history("d3", dax);
   call from_history("o3", oax);
   call from_history("n4", ny);
   call from_history("d4", dy);
   call from_history("o4", oy);

   !! Auxillary output files
   call from_param("outtag", out_);
   call from_Param("jsnap", jsnap, 50);


   ! Parallel Programming
   call from_param("nodes" ,nodes, 8)
   !nodes=8
   Totalnodes =omp_get_num_procs()
   call omp_set_num_threads(nodes); 
    write(0,*) 'OMP: using', nodes , 'out of', Totalnodes

   nhx = 64.
   dhx = 10
   ohx = -640.
   niter=10

   !*********************************
   write(0,*) '**************************'
   write(0,*) ' nz, dz, oz ', nz, dz, oz
   write(0,*) ' nx, dx, ox',  nx, dx, ox
   write(0,*) ' ny, dy, oy',  ny, dy, oy
   write(0,*) ' nax, dax, oax',  nax, dax, oax
   write(0,*) '**************************'

  !call erexit('exit now 11')

   !----------------------------------------------------------------

  !! output file (offset domain CIG)
   call auxputch("n1",'i', nz, trim(out_));      call auxputch("label1",'s', 'z', trim(out_)); 
   call auxputch("d1",'f', dz, trim(out_));      call auxputch("o1",'f', oz, trim(out_));
   call auxputch("n2",'i', nx, trim(out_));      call auxputch("label2",'s', 'x', trim(out_)); 
   call auxputch("d2",'f', dx, trim(out_));      call auxputch("o2",'f', ox, trim(out_));
   call auxputch("n3",'i', 2*nhx+1, trim(out_));  call auxputch("label3",'s', 'hx', trim(out_)); 
   call auxputch("d3",'f', dhx, trim(out_));      call auxputch("o3",'f', -nhx*dhx, trim(out_));

   !----------------------------------------------------------------
   ! Allocation

   write(0,*) 'allocating size', nx,ny,nz,1+2*nhx,nx*ny*nz*(1+2*nhx)
   allocate(model(nz,nx,ny,1+2*nax))
   allocate(model2(nz,nx,ny,1+2*nhx))
   allocate(tmpZX(nz,1+2*nhx), tmpZA(nz,1+2*nax), dealiasZA(nz,1+2*nax))
   allocate( tmpZA2(nz,1+2*nax), tmpZA3(nz,1+2*nax))

   !----------------------------------------------------------------
   ! initialize
   do iy = 1,ny
     do ix=1,nx
       do iz=1,nz
         model(iz,ix,iy,:) = 0.;
         model2(iz,ix,iy,:) = 0.;
       end do
     end do
   end do

   write(0,*) 'before reading ODCIG'
   ! Read in ODCIG
   call sep_read(model)

   ! Convert subsurface offset gather of ADCIG 
   ! using slant stacking
   call adcig_init(2*nhx+1,ohx,dhx,2*nax+1,oax,dax,nz,oz,dz)
   call Anglefilter_init(2*nax+1,oax,dax,2*nax+1,-2000.,40.,nz,oz,dz)
   call DealiasPTAU_init( 2*nhx+1,dhx,nz, dz)

   do iy=1,ny
     do ix=1,nx
       do iax = 1,1+2*nax
         do iz=1,nz   
           tmpZA(iz,iax) = model(iz,ix,iy,iax)
         end do
       end do
       call adcig2_oper(.true.,.false.,tmpZX,tmpZA)

       do ihx = 1,1+2*nhx
         do iz=1,nz   
            model2(iz,ix,iy,ihx) = tmpZX(iz,ihx)
         end do
       end do
     end do
   end do

   write(0,*) 'Output model'
   call sep_write(model2,out_,esize=4)   ! ADCIG


   call sep_close()
   deallocate(model,model2);


end program
 
