#include "blend.h"
#include "float_2d.h"

blend::blend(int btype, int *shifts_in, int nshots, int nts, hypercube_float *mod, hypercube_float *dat){
  //tag_init(tg);
  set_domain(mod);
  set_range(dat);

  shifts=shifts_in;  // Pointer to array of shot timings
  n_shots=nshots;    // Total number of shots
  nt_s=nts;          // Time length of each cut shot
fprintf(stderr,"Set up blend, shots %d, length %d, shifts",n_shots,nt_s);
for(int i1=0; i1<n_shots; i1++) fprintf(stderr," %d ",shifts[i1]);
fprintf(stderr,"\n");
}

//Take continuous data, cut and align
bool blend::adjoint(bool add, my_vector *model, my_vector *data){
  float_2d *d=(float_2d*) data;
  float_2d *m=(float_2d*) model;
  if(!add) m->zero();
  axis a1=d->get_axis(1);
  axis a2=d->get_axis(2);

  for(int i3=0; i3 < n_shots; i3++){  //shot
    for(int i2=0; i2 < a2.n; i2++){   //receiver
      for(int i1=0; i1 < nt_s; i1++){ //time
	it_s=shifts[i3];
        m->vals [i3*a1.n*a2.n+i2*a1.n+i1] += d->vals [it_s+i1+i2*a1.n] ;
      }
    }
  }
}

// Take cut data, make continuous
bool blend::forward(bool add, my_vector *model, my_vector *data){
  hypercube_float *d=(float_2d*) data;
  hypercube_float *m=(float_2d*) model;
  axis a1=d->get_axis(1);
  axis a2=d->get_axis(2);
  if(!add) d->zero();
fprintf(stderr,"In fwd \n");
  for(int i3=0; i3 < n_shots; i3++){
    for(int i2=0; i2 < a2.n; i2++){
      for(int i1=1; i1 < nt_s; i1++){
	it_s=shifts[i3];
	d->vals [it_s+i1+i2*a1.n] += m->vals [i3*a1.n*a2.n+i2*a1.n+i1];
      }
    }
  }
}

