#include "ext_image.h"
#include "math.h"

ext_image::ext_image(char *tg,vel_fd_2d *v,param_func *par){
  tagit=tg;

  nh=par->get_int("nh",1.);
  dh=par->get_float("dh",1.);
  oh=par->get_float("oh",-((nh-1)*dh)/2);
  fprintf(stderr,"Extended axis info %d %f %f \n",nh,dh,oh);
  a3=axis(nh,oh,dh);
  std::vector<axis> axes; axes.push_back(v->get_axis(1));
  axes.push_back(v->get_axis(2));
  axes.push_back(a3);
  image=new hypercube_float(axes);
  init_nd(axes);
  auxinout(tg);
  write_description(tg);
  image->zero();

  //mute=image->clone_zero();

}

void ext_image::set_source_file(oc_float *ptr){
     myf=ptr;
     sseek(myf->tagit.c_str(),0,0);
     read_all(myf->tagit,image);
}

hypercube_float *ext_image::create_img_domain(hypercube_float *dom){
  //ax.o=sx[ishot]-aper-ax.d*nbound;
  std::vector<axis> axes; axes.push_back(dom->get_axis(1));
  axes.push_back(dom->get_axis(2));
  axes.push_back(a3);
  hypercube_float *tmp=new hypercube_float(axes);
  return tmp; 
}

void ext_image::add_image(hypercube_float *img){
  axis a1o=get_axis(1);
  axis a2o=get_axis(2);
  axis a3o=get_axis(3);
  axis a1i=img->get_axis(1);
  axis a2i=img->get_axis(2);
  axis a3i=img->get_axis(3);

  if(a3o.n != a3i.n || a3o.o != a3i.o){
    fprintf(stderr,"Offset axis mismatch %d %d ; %f %f \n",a3o.n,a3i.n,a3o.o,a3i.o);
  }

  int f2=(int)((a2i.o-a2o.o)/a2o.d);
  int f1=(int)((a1i.o-a1o.o)/a1o.d);
  //fprintf(stderr,"Window out from %d in x and %d in z \n",f2,f1);
  for(int i3=0; i3 < a3i.n; i3++){
    for(int i2=0; i2 < a2i.n; i2++){
      if(i2+f2 >=0 && i2+f2 < a2o.n){
        for(int i1=0; i1 < a1i.n; i1++){
          if(i1+f1>=0 && i1+f1 <a1o.n){
             image->vals[i1+f1+(i2+f2)*a1o.n+i3*a1o.n*a2o.n]+=img->vals[i1+i2*a1i.n+i3*a1i.n*a2i.n];
          }
        }
      }
    }
  }

}


float_3d *ext_image::extract_sub(axis a1i, axis a2i){
  axis a1o=get_axis(1);
  axis a2o=get_axis(2);
  axis a3o=get_axis(3);

  sseek(myf->tagit.c_str(),0,0);
  read_all(myf->tagit,image);
  float_3d *img=new float_3d(a1i,a2i,a3o);
  img->zero();

  int f2=(int)((a2i.o-a2o.o)/a2o.d);
  int f1=(int)((a1i.o-a1o.o)/a1o.d);
  for(int i3=0; i3 < a3o.n; i3++){
    for(int i2=0; i2 < a2i.n; i2++){
      if(i2+f2 >=0 && i2+f2 < a2o.n){
        for(int i1=0; i1 < a1i.n; i1++){
          if(i1+f1>=0 && i1+f1 <a1o.n){
            img->vals[i1+i2*a1i.n+i3*a1i.n*a2i.n]=image->vals[i1+f1+(i2+f2)*a1o.n+i3*a1o.n*a2o.n];
          }
        }
      }
    }
  }
  return img;
}

void ext_image::write_volume(){
  sseek(myf->tagit.c_str(),0,0);
  write_all(myf->tagit,image);
}

//bool ext_image::adj_image(bool add, my_vector *model, my_vector *data, float *srcfld){
bool ext_image::adj_image(bool add, hypercube_float *model, float *data, float *srcfld){

  //float_3d *d=(float_3d*) data;
  /*float_3d *m=(float_3d*) model;
  src_fld=srcfld;
  if(!add) m->zero();

  int n1=d->get_axis(1).n;
  int n2=d->get_axis(2).n;
  int n3=d->get_axis(2).n;

  int nhx=(nh-1)/2;
  int ihx=0;

  for(int ih=0; ih<nh; ih++){
    ihx=ih-nhx;
    for(int it=0; it<n3; it++){
      for(int ix=0; ix<n2; ix++){
        for(int iz=0; iz<n1; iz++){
          if(ix > nhx && ix < (n1-nhx)){
            m->vals[iz+ix*n1+ih*n1*n2] += data->vals[iz+(ix+ihx)*n1+it*n1*n2]*src_fld[iz+(ix-ihx)*n1+it*n1*n2];
          }
	  else{
            m->vals[iz+ix*n1+ih*n1*n2] += data->vals[iz+ix*n1+it*n1*n2]*src_fld[iz+ix*n1+it*n1*n2];
          }
        }
      }
    }
  }*/
}

//bool ext_image::fwd_image(bool add, my_vector *model, my_vector *data, float *srcfld){
bool ext_image::fwd_image(bool add, hypercube_float *model, float *data, float *srcfld){
  //float_3d *d=(float_3d*) data;
  /*float_3d *m=(float_3d*) model;
  if(!add) d->zero();
  src_fld=srcfld;

  int n1=d->get_axis(1).n;
  int n2=d->get_axis(2).n;
  int n3=d->get_axis(2).n;

  int ihx=0;
  int nhx=(nh-1)/2;

  for(int ih=0; ih<nh; ih++){
    ihx=ih-nhx;
     for(int it=0; it<n3; it++){
       for(int iz=0; iz<n2; iz++){
         for(int ix=0; ix<n1; ix++){
	   if(ix > (nhx) && ix < (n1-nhx)){
            data->vals[iz+(ix+ihx)*n1+it*n1*n2] += m->vals[iz+ix*n1+ih*n1*n2]*src_fld[iz+(ix-ihx)*n1+it*n1*n2];
	   }
	   else{
            data->vals[iz+ix*n1+it*n1*n2] += m->vals[iz+ix*n1+ih*n1*n2]*src_fld[iz+ix*n1+it*n1*n2];
           }
	 }
       }
     }
   }*/

}

