#include "lin_solver.h"
#include "i_op.h"
#include "super_vector.h"
#include "combo_oper.h"

simple_solver::simple_solver(step *s, my_vector *mod, my_operator *op, my_operator *wt, my_vector *m0){
  init_solve(s);
  rr=init_rr(mod,op,wt,m0);
  //create_wt_op(op,wt);
  //oper=wt_op;
  oper=op;
  create_solver_vecs();
}

void lin_solver::init_solve(step *s){
  st=s;
  verb=0;
  have_wt=false;
  have_op=false;
}

void lin_solver::create_solver_vecs(){

   gg=rr->clone_zero();
   m=oper->range->clone_zero();//change from domain
   g=oper->range->clone_zero();
   st->alloc_step(m,rr);
}


my_vector *lin_solver::init_rr(my_vector *mod, my_operator *op, my_operator *wt, my_vector *m0){

 my_vector *rtemp;
 rtemp=mod->clone_vec();
 rtemp->scale(-1.);

 return rtemp;
}

bool lin_solver::solve(int niter){
double scale;
  if(verb>0){
    scale=rr->dot(rr);
    fprintf(stderr,"Initial residual %g scaled to 10000\n",scale);
    scale=100000/scale;
  }
  FILE *d_res;
  for(int iter =0; iter < niter; iter++){
 
    //oper->adjoint(false,g,rr);
    //oper->forward(false,g,gg);

    oper->forward(false,rr,g);
    oper->adjoint(false,gg,g);

    double val;

    bool valid=st->step_it(iter,m,g,rr,gg,&val);//iter,d_sp,d_sp,m_sp,m_sp
    m->info("mod_mov");
    rr->info("rr_mov");
    if(verb>0){
      fprintf(stderr,"Finished iteration: %d %g  \n",iter,(double)(val*scale));
    }

    d_res = fopen("residuals.H","a");
    fprintf(d_res,"Iteration %d, data space residual %g ; scaled %g \n",iter,val,(double)(val*scale));
    fclose(d_res);

    if(!valid) return false;

  }
  return true;
}

my_vector *lin_solver::return_model(){ return m->clone_vec();}

void lin_solver::update_model(my_vector *mod){mod->add(m);}

my_vector *prec_solver::return_model(){
  my_vector *mod=pop->range->clone_zero();
  pop->forward(false,m,mod);
 

  return mod;
}

void prec_solver::update_model(my_vector *mod){
  pop->forward(true,m,mod);
}

void prec_solver::clean_up(){
   delete iop;
   delete prec_chain_op;
}
void lin_solver::clean_up(){
   delete rr;
   delete m; delete gg; delete g; 
   if(have_wt) delete wt_op;
   if(have_op) delete oper;
}
void simple_solver::clean_up(){}
void reg_solver::clean_up(){}
void lin_solver::set_verbose(int iverb){ verb=iverb;}
