#include "rtm_zero_op.h"
#include "sinc_bob.h"
#include "gpu_funcs.h"
#include <math.h>
#include "deriv.h"
#include "map_data.h"
#include "weight.h"
#include "blend.h"

rtm_zero_op::rtm_zero_op(param_func *par,vel_fd_2d *vel_2d, source_func *source_func, data_rtm_2d *dat, ext_image *img,float ap,bool v,bool do_src,bool redo_src){
   std::vector<float> ds; 
   data=dat;
   ds.push_back(dat->dt);
   ds.push_back(source_func->get_dt());
   set_vel(vel_2d);
   calc_stability(&ds,data->nt);
   set_fd_basics(par,source_func,ap,v);
   image=img;
   set_domain(image);
   set_range(data);
   dtd=data->dt;
   jtd=(int)(dtd/dt);
   nt=(data->nt-1)*jtd+1;
   base=0;
   redo=redo_src;
   create_transfer_sinc_data(8);
   //std::cout<<"Redo is "<<redo<<" "<<redo_src<<std::endl;
  par_f=par;

   if(do_src && !redo) create_source_fields();
 
}

void rtm_zero_op::create_transfer_sinc_data(int nsinc){
  jtd=(int)((data->dt)/dt);
  sinc_bob myr(jtd,nsinc);
  transfer_sinc_table_d(nsinc,jtd,myr.table);
}


void rtm_zero_op::create_source_fields(){

  int nshots=data->nshots();
  auxinout("source_fields");  

  int nt_big=9+data->nt;
  int npts=source->get_points();
  int *locs=new int[npts];
  float *vals=new float[npts*(nt_big)];

  int *tempi=0;
  float *tempf=0;

  //nshots=1;
  for(int ishot=0; ishot< nshots; ishot++){
     if(verb) fprintf(stderr,"Forward propagating shot %d of %d \n",ishot,nshots);

     hypercube_float *src_p0=source->create_domain(ishot);
     hypercube_float *src_p1=src_p0->clone();
     hypercube_float *src_q0=src_p0->clone();
     hypercube_float *src_q1=src_p0->clone();
     hypercube_float *vloc=vel->rand_subspace(ishot+base,nboundt,nbound,dt,src_p0);

     source->get_source_func(src_p0,ishot,nt_big,locs,vals);
     transfer_source_func(npts,nt_big,locs,vals);
     transfer_vel_func1(nz,nx,vloc->vals);
     hypercube_float *vnone=vel->rand_subspace(ishot+base,0,0,dt,src_p0);
     transfer_vel_func2(nz,nx,vnone->vals);

     source_prop(nz,nx,false,true,src_p0->vals,src_p1->vals,jts,npts,nt+2,0,tempi,tempf,-1,"junk");

     srite("source_fields",src_p0->vals,nz*nx*4);
     srite("source_fields",src_p1->vals,nz*nx*4);
     delete src_q0;
     delete src_q1;
     delete src_p0;
     delete src_p1;
     delete vloc;
   }

   delete [] locs;
   delete [] vals;
}

bool rtm_zero_op::forward(bool add, my_vector *mvec, my_vector *dvec){
  data_rtm_2d *d=(data_rtm_2d*) dvec;
  ext_image *m=(ext_image*) mvec;
  if(!add) dvec->zero();

  int npts_s=source->get_points();

  data->set_source_file(d);
  image->set_source_file(m);

  int nt_big=8+data->nt;

  int npts=data->get_points();
  int *locs=new int[npts];

  int *locs_s=new int[npts_s];
  float *vals=new float[npts_s*(nt_big)];
  //float *s_z=new float[npts/9],*s_x=new float[npts/9];
  float *s_z=new float[npts],*s_x=new float[npts];

  axis ax_pt(npts);
  float_2d *shot=new float_2d(data->get_axis(1),data->get_axis(2));
  float_2d *shot_deriv=new float_2d(data->get_axis(1),data->get_axis(2));

  axis ax_t=axis(nt_big);
  float_2d *rec_func=new float_2d(ax_t,ax_pt);

  int nshots=data->nshots();
  int nts=(int)(nt/jts)+1;
  //nshots=1;
  for(int ishot=0; ishot<nshots; ishot++){

    if(verb) fprintf(stderr,"      Forward shot %d of %d \n",ishot,data->nshots());
    hypercube_float *src_p0=source->create_domain(ishot);
    hypercube_float *src_p1=src_p0->clone();
    axis ad1=src_p0->get_axis(1);
    axis ad2=src_p0->get_axis(2);

    hypercube_float *vrand=vel->rand_subspace(ishot+base,nboundt,nbound,dt,src_p0);
    hypercube_float *vnone=vel->rand_subspace(ishot+base,0,0,dt,src_p0);
    hypercube_float *img=image->extract_sub(ad1,ad2);
    transfer_vel_func1(nz,nx,vrand->vals);
    transfer_vel_func2(nz,nx,vnone->vals);
    source->get_source_func(src_p0,ishot,nt_big,locs_s,vals);
    transfer_source_func(npts_s,nt_big,locs_s,vals);   

    data->get_source_func(src_p0,ishot,s_z,s_x,8,nt_big,shot);
    map_data mapit(npts,s_z,s_x,locs,src_p0,rec_func,shot_deriv,nt_big);

    transfer_receiver_func(npts,nt_big,locs,rec_func->vals);

    rtm_forward(ad1.n,ad2.n,jtd,img->vals,rec_func->vals,npts_s,npts,nt,nt_big,npts,image->get_axis(3).n);

    deriv der=deriv(shot_deriv,shot);
    mapit.forward(false,rec_func,shot_deriv);
    der.forward(false,shot_deriv,shot);
    data->add_data(ishot,shot);

    delete src_p0; delete src_p1; delete img;
    delete vrand; delete vnone;
  }

  delete [] s_z;
  delete [] s_x;
  delete [] locs_s; delete [] vals;
  //delete [] shot; delete [] shot_deriv; delete [] rec_func;
  return true;
   
}

bool rtm_zero_op::adjoint(bool add, my_vector *mvec, my_vector *dvec){

  if(redo) create_source_fields();
  if(!add) mvec->zero();

  data_rtm_2d *d=(data_rtm_2d*) dvec;
  ext_image *m=(ext_image*) mvec;

  int npts_s=source->get_points();
  data->set_source_file(d);

  image->set_source_file(m);
  int nt_big=8+data->nt;
  int npts=data->get_points();
  int *locs=new int[npts];
  int *locs_s=new int[npts_s];
  float *vals=new float[npts_s*(nt_big)];

  float *s_z=new float[npts],*s_x=new float[npts];
  axis ax_pt(npts);

  float_2d *shot=new float_2d(data->get_axis(1),data->get_axis(2));
  float_2d *shot_deriv=new float_2d(data->get_axis(1),data->get_axis(2));
  axis ax_t=axis(nt_big);
  float_2d *rec_func=new float_2d(ax_t,ax_pt);
  sseek("source_fields",0,0);
  image->zero();
  int nts=(int)(nt/jts)+1;

  int nshots=data->nshots();
  //nshots=1;
  float *srcfld,*recfld;

  for(int ishot=0; ishot<nshots; ishot++){
    if(verb) fprintf(stderr,"      Adjoint shot %d of %d \n",ishot,data->nshots());
    hypercube_float *src_p0=source->create_domain(ishot);
    hypercube_float *src_p1=src_p0->clone();
    axis ad1=src_p0->get_axis(1);
    axis ad2=src_p0->get_axis(2);
    hypercube_float *vrand=vel->rand_subspace(ishot+base,nboundt,nbound,dt,src_p0);
    hypercube_float *vnone=vel->rand_subspace(ishot+base,0,0,dt,src_p0);

    hypercube_float *img=image->create_img_domain(src_p0);

    transfer_vel_func1(nz,nx,vrand->vals);
    transfer_vel_func2(nz,nx,vnone->vals);

    sreed("source_fields",src_p1->vals,4*ad1.n*ad2.n);
    sreed("source_fields",src_p0->vals,4*ad1.n*ad2.n);
    source->get_source_func(src_p0,ishot,nt_big,locs_s,vals);
    transfer_source_func(npts_s,nt_big,locs_s,vals); 

    data->get_source_func(src_p0,ishot,s_z,s_x,8,nt_big,shot);



    deriv der=deriv(shot_deriv,shot);
    der.adjoint(false,shot_deriv,shot);

    map_data mapit(npts,s_z,s_x,locs,src_p0,rec_func,shot_deriv,nt_big);
    mapit.adjoint(false,rec_func,shot_deriv);

    transfer_receiver_func(npts,nt_big,locs,rec_func->vals);

    rtm_adjoint(ad1.n,ad2.n,jtd,src_p0->vals,src_p1->vals,img->vals,srcfld,recfld,npts_s,npts,nt,image->get_axis(3).n);

    weight w(img,300,1);
    w.adjoint(true,img,img);

    /*float z=0,h=0,z0=550,h0=1;
    float hmax=image->get_axis(3).o+image->get_axis(3).d*image->get_axis(3).n;

    for(int i3=0;i3<image->get_axis(3).n;i3++){
     h=image->get_axis(3).o+i3*image->get_axis(3).d;
     for(int i2=0;i2<img->get_axis(2).n;i2++){
      for(int i1=0;i1<img->get_axis(1).n;i1++){
        z=z0+img->get_axis(1).o+i1*img->get_axis(1).d;
        if(z<0) z=0;
        img->vals[i1+i2*img->get_axis(1).n+i3*img->get_axis(1).n*img->get_axis(2).n]*=(z*(hmax-abs(h)+h0));
      }
     }
    }*/

    image->add_image(img);
    delete src_p0; delete src_p1; delete img;
    delete vrand; delete vnone;
  }

  image->write_volume();

  delete [] s_z;
  delete [] s_x;
  delete [] locs_s; delete [] vals;

  return true;
}
