/* 2D Time domain finite difference kernel
	By: Abdullah AlTheyab
	Ref: similar to Paulius code but for 2D
 */
#define BLOCK_SIZE 16
#define FAT 4
#define COEFFS_SIZE 11

#define C0  0
#define CZ1 1
#define CX1 2
#define CZ2 3
#define CX2 4
#define CZ3 5
#define CX3 6
#define CZ4 7
#define CX4 8
#define C0Z 9
#define C0X 10

#define C_C00(d) (8.0/(5.0*(d)*(d)))
#define get_coeffs(d1,d2) {-1025.0/576.0*(C_C00(d1)+C_C00(d2)), C_C00(d1), C_C00(d2),-C_C00(d1)/8.0,-C_C00(d2)/8.0,C_C00(d1)/63.0,C_C00(d2)/63.0,-C_C00(d1)/896.0,-C_C00(d2)/896.0,-1025.0/576.0*C_C00(d1),-1025.0/576.0*C_C00(d2)}
__constant__ float coeffs[COEFFS_SIZE];


extern "C" __global__ void wave_kernel_adj(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
     __shared__ float vls[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];

	int ig=FAT+blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global z coordinates
	int jg=FAT+blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates
	int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	int addr=dloc(ig,jg);	// linear memroy address
	//copy to shared memory
	p1s[jl][il]=p1[addr];
	vls[jl][il]=tex2D(vel_ref2,ig,jg);

	if(threadIdx.y<FAT){
		p1s[jl-FAT][il]=p1[dloc(ig,jg-FAT)];
		p1s[jl+BLOCK_SIZE][il]=p1[dloc(ig,jg+BLOCK_SIZE)];
		vls[jl-FAT][il]=tex2D(vel_ref2,ig,jg-FAT);
		vls[jl+BLOCK_SIZE][il]=tex2D(vel_ref2,ig,jg+BLOCK_SIZE);
	}
	if(threadIdx.x<FAT){
		p1s[jl][il-FAT]=p1[dloc(ig-FAT, jg)];
		p1s[jl][il+BLOCK_SIZE]=p1[dloc(ig+BLOCK_SIZE, jg)];
	    vls[jl][il-FAT]=tex2D(vel_ref2,ig-FAT,jg);
		vls[jl][il+BLOCK_SIZE]=tex2D(vel_ref2,ig+BLOCK_SIZE,jg);
	}
		__syncthreads();
			//p2[addr]= p1[addr]+p0[addr];
			



//	p2[addr]=  //tex2D(vel_ref1, ig,jg)*
/*
        p2[addr]=tex2D(vel_ref1,ig,jg)*coeffs[C0]*p1s[jl][il]+
                 tex2D(vel_ref1,ig,jg+1)*coeffs[CX1]*p1s[jl+1][il]+
                 tex2D(vel_ref1,ig,jg-1)*coeffs[CX1]*p1s[jl-1][il]+
                 tex2D(vel_ref1,ig,jg+2)*coeffs[CX2]*p1s[jl+2][il]+
                 tex2D(vel_ref1,ig,jg-2)*coeffs[CX2]*p1s[jl-2][il]+
                 tex2D(vel_ref1,ig,jg+3)*coeffs[CX3]*p1s[jl+3][il]+
                 tex2D(vel_ref1,ig,jg-3)*coeffs[CX3]*p1s[jl-3][il]+
                 tex2D(vel_ref1,ig,jg+4)*coeffs[CX4]*p1s[jl+4][il]+
                 tex2D(vel_ref1,ig,jg-4)*coeffs[CX4]*p1s[jl-4][il]+
                 tex2D(vel_ref1,ig+1,jg)*coeffs[CZ1]*p1s[jl][il+1]+
                 tex2D(vel_ref1,ig-1,jg)*coeffs[CZ1]*p1s[jl][il-1]+
                 tex2D(vel_ref1,ig+2,jg)*coeffs[CZ2]*p1s[jl][il+2]+
                 tex2D(vel_ref1,ig-2,jg)*coeffs[CZ2]*p1s[jl][il-2]+
                 tex2D(vel_ref1,ig+3,jg)*coeffs[CZ3]*p1s[jl][il+3]+
                 tex2D(vel_ref1,ig-3,jg)*coeffs[CZ3]*p1s[jl][il-3]+
                 tex2D(vel_ref1,ig+4,jg)*coeffs[CZ4]*p1s[jl][il+4]+
                 tex2D(vel_ref1,ig-4,jg)*coeffs[CZ4]*p1s[jl][il-4]
                 */
                 
        p2[addr]=vls[jl][il]*coeffs[C0]*p1s[jl][il]+
                 vls[jl+1][il]*coeffs[CX1]*p1s[jl+1][il]+
                 vls[jl-1][il]*coeffs[CX1]*p1s[jl-1][il]+
                 vls[jl+2][il]*coeffs[CX2]*p1s[jl+2][il]+
                 vls[jl-2][il]*coeffs[CX2]*p1s[jl-2][il]+
                 vls[jl+3][il]*coeffs[CX3]*p1s[jl+3][il]+
                 vls[jl-3][il]*coeffs[CX3]*p1s[jl-3][il]+
                 vls[jl+4][il]*coeffs[CX4]*p1s[jl+4][il]+
                 vls[jl-4][il]*coeffs[CX4]*p1s[jl-4][il]+
                 vls[jl][il+1]*coeffs[CZ1]*p1s[jl][il+1]+
                 vls[jl][il-1]*coeffs[CZ1]*p1s[jl][il-1]+
                 vls[jl][il+2]*coeffs[CZ2]*p1s[jl][il+2]+
                 vls[jl][il-2]*coeffs[CZ2]*p1s[jl][il-2]+
                 vls[jl][il+3]*coeffs[CZ3]*p1s[jl][il+3]+
                 vls[jl][il-3]*coeffs[CZ3]*p1s[jl][il-3]+
                 vls[jl][il+4]*coeffs[CZ4]*p1s[jl][il+4]+
                 vls[jl][il-4]*coeffs[CZ4]*p1s[jl][il-4]
                 
/*
		p2[addr]=  tex2D(vel_ref1, ig,jg)*

		( 0+
	//		coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
	//		+coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
	//		+coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
	//		+coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
	// 		+coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
	//		+coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
	//		+coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
		  +coeffs[C0]*p1s[jl][il]
	)
	*/
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			

}
extern "C" __global__ void wave_kernel2(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];

	int ig=FAT+blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global z coordinates
	int jg=FAT+blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates
	int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	int addr=dloc(ig,jg);	// linear memroy address
	//copy to shared memory
	p1s[jl][il]=p1[addr];
    
	if(threadIdx.y<FAT){
		p1s[jl-FAT][il]=p1[dloc(ig,jg-FAT)];
		p1s[jl+BLOCK_SIZE][il]=p1[dloc(ig,jg+BLOCK_SIZE)];
	}
	if(threadIdx.x<FAT){
		p1s[jl][il-FAT]=p1[dloc(ig-FAT, jg)];
		p1s[jl][il+BLOCK_SIZE]=p1[dloc(ig+BLOCK_SIZE, jg)];
	}
		__syncthreads();


// fprintf(stderr,"%g %g %g %g %g \n",coeffs[0],coeffs[1],coeffs[2],coeffs[3],coeffs[4]);
	//	p2[addr]=  tex2D(vel_ref1, ig,jg)*


  p2[addr]= tex2D(vel_ref2, ig,jg)*

		( 0+
			coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
			+coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
			+coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
			+coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
	 		+coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
			+coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
		+coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
		  +coeffs[C0]*p1s[jl][il]
	)
			+p1s[jl][il]+ p1s[jl][il]-p0[addr];
}
extern "C" __global__ void wave_kernel(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];

	int ig=FAT+blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global z coordinates
	int jg=FAT+blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates
	int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	int addr=dloc(ig,jg);	// linear memroy address
	//copy to shared memory
	p1s[jl][il]=p1[addr];
    
	if(threadIdx.y<FAT){
		p1s[jl-FAT][il]=p1[dloc(ig,jg-FAT)];
		p1s[jl+BLOCK_SIZE][il]=p1[dloc(ig,jg+BLOCK_SIZE)];
	}
	if(threadIdx.x<FAT){
		p1s[jl][il-FAT]=p1[dloc(ig-FAT, jg)];
		p1s[jl][il+BLOCK_SIZE]=p1[dloc(ig+BLOCK_SIZE, jg)];
	}
		__syncthreads();


// fprintf(stderr,"%g %g %g %g %g \n",coeffs[0],coeffs[1],coeffs[2],coeffs[3],coeffs[4]);
	//	p2[addr]=  tex2D(vel_ref1, ig,jg)*


  p2[addr]= tex2D(vel_ref1, ig,jg)*

		( 0+
			coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
			+coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
			+coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
			+coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
	 		+coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
			+coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
		+coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
		  +coeffs[C0]*p1s[jl][il]
	)
			+p1s[jl][il]+ p1s[jl][il]-p0[addr];

}
extern "C" __global__ void damp_kernel(float *p0, float *p1){
	
	int ig=FAT+blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global z coordinates
	int jg=FAT+blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates
  int edge=min4(ig-FAT,jg-FAT,n2gpu-FAT-jg,n1gpu-FAT-ig);
	if(edge>=0){
	int addr=dloc(ig,jg);		  		  			// linear memroy address
	float temp=expf(-bc_bgpu*(bc_agpu-edge));
		if(temp<1.){

  p1[addr]*=temp;
  p0[addr]*=temp;
  }
  }
 // if(ig==10 && jg==10) fprintf(stderr,"CHECK THIS %f %f %f \n",bc_agpu,bc_bgpu,temp);

}

extern "C" __global__ void wave_kernel_damp(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
	int ig=FAT+blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global z coordinates
	int jg=FAT+blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates
	int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	int addr=dloc(ig,jg);		  		  			// linear memroy address

	//copy to shared memory
	p1s[jl][il]=p1[addr];

	if(threadIdx.y<FAT){
		p1s[jl-FAT][il]=p1[dloc(ig,jg-FAT)];
		p1s[jl+BLOCK_SIZE][il]=p1[dloc(ig,jg+BLOCK_SIZE)];
	}
	if(threadIdx.x<FAT){
		p1s[jl][il-FAT]=p1[dloc(ig-FAT, jg)];
		p1s[jl][il+BLOCK_SIZE]=p1[dloc(ig+BLOCK_SIZE, jg)];
	}
	
	__syncthreads();
    int edge=min4(ig,jg,n2gpu-jg,n1gpu-ig);
	float temp=expf(-bc_bgpu*(bc_agpu-edge));
	if(temp>1.) temp=1.;
	//if(jg==20 && ig==20) fprintf(stderr,"EXP IS %f \n",temp);

      p2[addr]= 
       tex2D(vel_ref2, ig,jg)*
       (
			coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
			+coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
			+coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
			+coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
			+coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
			+coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
			+coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
			+coeffs[C0]*p1s[jl][il])-p0[addr]+p1[addr]+
			p1[addr];
			p1[addr]*=temp;
			p2[addr]*=temp;
	//		p1[addr]=p1[addr]*temp;
	//if(ig==104 && jg==1585) fprintf(stderr,"CHECK VAL %f %f %f %f \n",
//	   temp,  tex2D(vel_ref2, ig,jg),p1[addr],p2[addr]);

/*
	p2[addr]= tex2D(vel_ref2, ig,jg)*(
			coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
			+coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
			+coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
			+coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
			+coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
			+coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
			+coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
			-temp*p0[addr] 
			+(temp+coeffs[C0])*p1s[jl][il]
			)+p1s[jl][il]+p1s[jl][il]-p0[addr];
		
	*/		
}


