#define BLOCK_SIZE 12
#define FAT 2
#define COEFFS_SIZE 5

#define C0  0
#define CZ1 1
#define CX1 2
#define CZ2 3
#define CX2 4

#define C_C00(d) 1.0/(12.0*(d)*(d))
#define get_coeffs(d1,d2) {-30.0*(C_C00(d1)+C_C00(d2)), 16.0*C_C00(d1),16.0*C_C00(d2),-C_C00(d1),-C_C00(d2)}
__constant__ float coeffs[COEFFS_SIZE];


__global__ void wave_kernel(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];

	register int ig=blockIdx.x*BLOCK_SIZE+threadIdx.x; // global zdir
	register int jg=blockIdx.y*BLOCK_SIZE+threadIdx.y; // global xdir
	register int addr=dloc(ig,jg);

	//copy to shared memory
	p1s[threadIdx.y][threadIdx.x]=p1[addr];

	//Guard against fat threads
	if(threadIdx.x<FAT || threadIdx.x>=BLOCK_SIZE+FAT|| threadIdx.y<FAT || threadIdx.y>=BLOCK_SIZE+FAT) return; 

	__syncthreads();

	float temp= -expf(bc_bgpu+min3(jg,n2gpu-jg,n1gpu-ig)*bc_agpu); 
	
	p2[addr]=  tex2D(vel_ref, ig,jg)*(
			 coeffs[CX2]*(p1s[threadIdx.y-2][threadIdx.x]+p1s[threadIdx.y+2][threadIdx.x])
			+coeffs[CZ2]*(p1s[threadIdx.y][threadIdx.x+2]+p1s[threadIdx.y][threadIdx.x-2])
			+coeffs[CX1]*(p1s[threadIdx.y-1][threadIdx.x]+p1s[threadIdx.y+1][threadIdx.x])
			+coeffs[CZ1]*(p1s[threadIdx.y][threadIdx.x+1]+p1s[threadIdx.y][threadIdx.x-1])
			-temp*p0[addr] 
			+(temp+coeffs[C0])*p1s[threadIdx.y][threadIdx.x]
			)+2.f*p1s[threadIdx.y][threadIdx.x]-p0[addr];
}
